// Code generated by "stringer -type KeyType"; DO NOT EDIT.

package gokey

import "strconv"

const _KeyType_name = "EC256EC521RSA2048RSA4096X25519ED25519"

var _KeyType_index = [...]uint8{0, 5, 10, 17, 24, 30, 37}

func (i KeyType) String() string {
	if i < 0 || i >= KeyType(len(_KeyType_index)-1) {
		return "KeyType(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _KeyType_name[_KeyType_index[i]:_KeyType_index[i+1]]
}
